
TEXT = 'CharField'
BOOLEAN = 'BooleanField'
INTEGER = 'IntegerField'
FLOAT = 'FloatField'
DATE_TIME = 'DateTimeField'
TIMESTAMP = 'TimeStampField'
FOREIGN_KEY = 'ForeignKey'
JSON = 'JSONField'
GEOMETRY = 'GeometryField'
GEOGRAPHY = 'GeographyField'



from sqlalchemy import desc, case
from sqlalchemy.orm import load_only

from jet_bridge import fields
from jet_bridge.serializers.serializer import Serializer


def get_reset_order_serializer(Model, queryset, session):
    class ResetOrderSerializer(Serializer):
        ordering_field = fields.CharField()
        ordering = fields.CharField(required=False, allow_null=True)
        value_ordering = fields.CharField(required=False, allow_null=True)

        def save(self):
            ordering_field = self.validated_data['ordering_field']
            ordering = self.validated_data.get('ordering')
            value_ordering = self.validated_data.get('value_ordering')

            qs = queryset
            order_by = []

            if value_ordering:
                field, values_str = value_ordering.split('-', 2)
                values = values_str.split(',')
                order_by.append(case(
                    [(getattr(Model, field) == x, i) for i, x in enumerate(values)],
                    else_=len(values)
                ))

            if ordering:
                def map_field(name):
                    descending = False
                    if name.startswith('-'):
                        name = name[1:]
                        descending = True
                    field = getattr(Model, name)
                    if descending:
                        field = desc(field)
                    return field

                order_by.extend(map(lambda x: map_field(x), ordering.split(',')))

            if len(order_by):
                qs = qs.order_by(*order_by)

            i = 1

            for instance in qs.options(load_only(ordering_field)):
                setattr(instance, ordering_field, i)
                i += 1

            session.commit()

    return ResetOrderSerializer



from jet_bridge.views.base.generic_api import GenericAPIView
from jet_bridge.views.mixins.create import CreateAPIViewMixin
from jet_bridge.views.mixins.destroy import DestroyAPIViewMixin
from jet_bridge.views.mixins.list import ListAPIViewMixin
from jet_bridge.views.mixins.retrieve import RetrieveAPIViewMixin
from jet_bridge.views.mixins.update import UpdateAPIViewMixin


class ModelAPIViewMixin(ListAPIViewMixin,
                        RetrieveAPIViewMixin,
                        DestroyAPIViewMixin,
                        CreateAPIViewMixin,
                        UpdateAPIViewMixin,
                        GenericAPIView):
    pass



from sqlalchemy import text
from sqlalchemy.exc import SQLAlchemyError

from jet_bridge import fields
from jet_bridge.db import Session
from jet_bridge.exceptions.sql import SqlError
from jet_bridge.exceptions.validation_error import ValidationError
from jet_bridge.fields.sql_params import SqlParamsSerializers
from jet_bridge.serializers.serializer import Serializer


class SqlSerializer(Serializer):
    query = fields.CharField()
    params = SqlParamsSerializers(required=False)

    class Meta:
        fields = (
            'query',
            'params',
        )

    def validate_query(self, value):
        forbidden = ['insert', 'update', 'delete', 'grant', 'show']
        for i in range(len(forbidden)):
            forbidden.append('({}'.format(forbidden[i]))
        if any(map(lambda x: ' {} '.format(value.lower()).find(' {} '.format(x)) != -1, forbidden)):
            raise ValidationError('forbidden query')

        i = 0
        while value.find('%s') != -1:
            value = value.replace('%s', ':param_{}'.format(i), 1)
            i += 1

        return value

    def execute(self):
        session = Session()

        query = self.validated_data['query']
        params = self.validated_data.get('params', [])

        try:
            result = session.execute(
                text(query),
                params
            )

            rows = list(map(lambda x: x.itervalues(), result))

            def map_column(x):
                if x == '?column?':
                    return
                return x

            return {'data': rows, 'columns': map(map_column, result.keys())}
        except (SQLAlchemyError, TypeError) as e:
            raise SqlError(e)
        finally:
            session.close()



from jet_bridge import status
from jet_bridge.exceptions.api import APIException


class ValidationError(APIException):
    default_detail = 'Invalid input'
    default_code = 'invalid'
    default_status_code = status.HTTP_400_BAD_REQUEST



from sqlalchemy import Column, Integer, Sequence, String, DateTime

from jet_bridge.models import Base


class Token(Base):
    __tablename__ = '__jet__token'
    id = Column(Integer, Sequence('token_id_seq'), primary_key=True)
    project = Column(String(64))
    token = Column(String(32))
    date_add = Column(DateTime(timezone=True))

    def __repr__(self):
        return '<Token(project=\'%s\', token=\'%s\', date_add=\'%s\')>' % (self.project, self.token, self.date_add)



from jet_bridge.filters.filter_class import FilterClass
from jet_bridge.filters.model_m2m import get_model_m2m_filter
from jet_bridge.filters.model_search import get_model_search_filter
from jet_bridge.filters.model_segment import get_model_segment_filter
from jet_bridge.filters.order_by import OrderFilter


def get_model_filter_class(Model):
    search_filter = get_model_search_filter(Model)
    model_m2m_filter = get_model_m2m_filter(Model)
    model_segment_filter = get_model_segment_filter(Model)

    class ModelFilterClass(FilterClass):
        _order_by = OrderFilter()
        _search = search_filter()
        _m2m = model_m2m_filter()
        _segment = model_segment_filter()

        class Meta:
            model = Model

    return ModelFilterClass



from jet_bridge.responses.base import Response


class RetrieveAPIViewMixin(object):

    def retrieve(self, *args, **kwargs):
        instance = self.get_object()
        serializer = self.get_serializer(instance=instance)
        self.write_response(Response(serializer.representation_data))



from sqlalchemy import inspect

from jet_bridge import fields
from jet_bridge.serializers.model_serializer import ModelSerializer

data_types = [
    {'query': 'VARCHAR', 'operator': 'startswith', 'date_type': fields.CharField},
    {'query': 'TEXT', 'operator': 'equals', 'date_type': fields.CharField},
    {'query': 'BOOLEAN', 'operator': 'equals', 'date_type': fields.BooleanField},
    {'query': 'INTEGER', 'operator': 'equals', 'date_type': fields.CharField},
    {'query': 'SMALLINT', 'operator': 'equals', 'date_type': fields.CharField},
    {'query': 'NUMERIC', 'operator': 'startswith', 'date_type': fields.CharField},
    {'query': 'VARCHAR', 'operator': 'startswith', 'date_type': fields.CharField},
    {'query': 'TIMESTAMP', 'operator': 'startswith', 'date_type': fields.CharField},
    {'query': 'INTEGER[]', 'operator': 'startswith', 'date_type': fields.ArrayField},
    {'query': 'JSON', 'operator': 'startswith', 'date_type': fields.JSONField},
    {'query': 'geometry', 'operator': 'startswith', 'date_type': fields.WKTField},
    {'query': 'geography', 'operator': 'startswith', 'date_type': fields.WKTField},
]
default_data_type = fields.CharField


def get_column_data_type(column):
    try:
        data_type = str(column.type)
    except:
        data_type = 'NullType'

    for rule in data_types:
        if rule['operator'] == 'equals' and data_type == rule['query']:
            return rule['date_type']
        elif rule['operator'] == 'startswith' and data_type[:len(rule['query'])] == rule['query']:
            return rule['date_type']

    return default_data_type


def map_column(column):
    date_type = get_column_data_type(column)
    kwargs = {}
    if column.primary_key:
        kwargs['read_only'] = True
    return (column.key, date_type(**kwargs))


def get_model_serializer(Model):
    mapper = inspect(Model)

    class CustomModelSerializer(ModelSerializer):
        class Meta:
            model = Model
            dynamic_fields = dict(map(map_column, mapper.columns))

    return CustomModelSerializer



import os

import tornado.ioloop
import tornado.web

from jet_bridge import settings, media
from jet_bridge.router import Router
from jet_bridge.views.api import ApiHandler
from jet_bridge.views.image_resize import ImageResizeHandler
from jet_bridge.views.media_file_upload import MediaFileUploadHandler
from jet_bridge.views.main import MainHandler
from jet_bridge.views.message import MessageHandler
from jet_bridge.views.model import ModelHandler
from jet_bridge.views.model_description import ModelDescriptionsHandler
from jet_bridge.views.not_found import NotFoundHandler
from jet_bridge.views.register import RegisterHandler
from jet_bridge.views.sql import SqlHandler


def make_app():
    router = Router()

    router.register('/api/models/(?P<model>[^/]+)/', ModelHandler)

    urls = [
        (r'/', MainHandler),
        (r'/register/', RegisterHandler),
        (r'/api/', ApiHandler),
        (r'/api/register/', RegisterHandler),
        (r'/api/model_descriptions/', ModelDescriptionsHandler),
        (r'/api/sql/', SqlHandler),
        (r'/api/messages/', MessageHandler),
        (r'/api/file_upload/', MediaFileUploadHandler),
        (r'/api/image_resize/', ImageResizeHandler),
        (r'/media/(.*)', tornado.web.StaticFileHandler, {'path': settings.MEDIA_ROOT}),
    ]
    urls += router.urls

    if settings.MEDIA_STORAGE == media.MEDIA_STORAGE_DEFAULT:
        urls.append((r'/media/(.*)', tornado.web.StaticFileHandler, {'path': settings.MEDIA_ROOT}))

    return tornado.web.Application(
        handlers=urls,
        debug=settings.DEBUG,
        default_handler_class=NotFoundHandler,
        template_path=os.path.join(settings.BASE_DIR, 'templates')
    )



from sqlalchemy import inspect, desc, func
from sqlalchemy.orm import load_only
from sqlalchemy.sql import operators
from sqlalchemy.sql.elements import AnnotatedColumnElement, UnaryExpression


def get_row_number(Model, queryset, instance):
    mapper = inspect(Model)
    pk = mapper.primary_key[0].name

    subqquery = queryset.with_entities(
        mapper.primary_key[0].label('__inner__pk'),
        func.row_number().over(order_by=queryset._order_by).label('__inner__row')
    ).subquery()

    rest = queryset.session.query(subqquery.c.__inner__row).filter(subqquery.c.__inner__pk == getattr(instance, pk))
    return rest.scalar()


def get_row_siblings(Model, queryset, row_number):
    mapper = inspect(Model)
    pk = mapper.primary_key[0].name

    has_prev = row_number > 1
    offset = row_number - 2 if has_prev else row_number - 1
    limit = 3 if has_prev else 2

    rows = queryset.options(load_only(pk)).limit(limit).offset(offset).all()

    if has_prev:
        next_index = 2
    else:
        next_index = 1

    if next_index >= len(rows):
        next_index = None

    if has_prev:
        prev_index = 0
    else:
        prev_index = None

    def map_row(row):
        return dict(((pk, getattr(row, pk)),))

    return {
        'prev': map_row(rows[prev_index]) if prev_index is not None else None,
        'next': map_row(rows[next_index]) if next_index is not None else None
    }


def get_model_siblings(Model, instance, queryset):
    mapper = inspect(Model)
    pk = mapper.primary_key[0].name
    context = queryset._compile_context()
    ordering = context.order_by

    def is_pk(x):
        if isinstance(x, AnnotatedColumnElement):
            return x.name == pk
        elif isinstance(x, UnaryExpression):
            return x.element.name == pk and x.modifier == operators.desc_op
        return False

    if ordering is None or not any(map(is_pk, ordering)):
        order_by = list(ordering or []) + [desc(pk)]
        queryset = queryset.order_by(*order_by)

    row_number = get_row_number(Model, queryset, instance)

    if not row_number:
        return {}

    return get_row_siblings(Model, queryset, row_number)



from sqlalchemy import sql, desc

from jet_bridge.filters.char_filter import CharFilter
from jet_bridge.filters.filter import EMPTY_VALUES


class OrderFilter(CharFilter):

    def filter(self, qs, value):
        if value in EMPTY_VALUES:
            return qs

        if len(value) < 2:
            return qs.filter(sql.false())

        ordering = value.split(',')

        def map_field(name):
            descending = False
            if name.startswith('-'):
                name = name[1:]
                descending = True
            field = getattr(self.model, name)
            if descending:
                field = desc(field)
            return field

        if len(ordering):
            qs = qs.order_by(*map(lambda x: map_field(x), ordering))

        return qs



from jet_bridge import fields
from jet_bridge.serializers.serializer import Serializer


class ModelDescriptionFieldSerializer(Serializer):
    name = fields.CharField()
    db_column = fields.CharField()
    field = fields.CharField()
    filterable = fields.BooleanField()
    editable = fields.BooleanField()
    params = fields.JSONField()

    class Meta:
        fields = (
            'name',
            'db_column',
            'field',
            'filterable',
            'editable',
            'params',
        )


class ModelDescriptionRelationSerializer(Serializer):
    name = fields.CharField()
    related_model = fields.JSONField()
    field = fields.CharField()
    related_model_field = fields.CharField()
    through = fields.JSONField()

    class Meta:
        fields = (
            'name',
            'related_model',
            'field',
            'related_model_field',
            'through',
        )


class ModelDescriptionSerializer(Serializer):
    model = fields.CharField()
    db_table = fields.CharField()
    hidden = fields.BooleanField()
    primary_key_field = fields.CharField()
    fields = ModelDescriptionFieldSerializer(many=True)
    relations = ModelDescriptionRelationSerializer(many=True)

    class Meta:
        fields = (
            'model',
            'db_table',
            'hidden',
            'fields',
            'relations',
            'primary_key_field',
        )




class Pagination(object):
    count = None

    def paginate_queryset(self, queryset, handler):
        raise NotImplementedError

    def get_paginated_response(self, data):
        raise NotImplementedError



from jet_bridge import status
from jet_bridge.exceptions.api import APIException


class PermissionDenied(APIException):
    default_detail = 'Permission denied'
    default_code = 'permission_denied'
    default_status_code = status.HTTP_403_FORBIDDEN



import json

from jet_bridge.fields.field import Field


class JSONField(Field):

    def to_internal_value_item(self, value):
        return json.loads(value)

    def to_representation_item(self, value):
        return json.dumps(value)



import dateparser
import six

from jet_bridge.fields.field import Field


class DateTimeField(Field):

    def to_internal_value_item(self, value):
        if value is None:
            return
        value = six.text_type(value)
        return dateparser.parse(value.strip())

    def to_representation_item(self, value):
        if value is None:
            return
        return value.isoformat()



from collections import OrderedDict, Mapping, Iterable

from jet_bridge.exceptions.validation_error import ValidationError
from jet_bridge.fields.field import Field
from jet_bridge.filters.filter import EMPTY_VALUES


class Serializer(Field):
    validated_data = None
    fields = []
    errors = None

    def __init__(self, *args, **kwargs):
        self.instance = kwargs.pop('instance', None)
        self.data = kwargs.pop('data', None)
        self.meta = getattr(self, 'Meta', None)
        self.partial = kwargs.pop('partial', False)
        super(Serializer, self).__init__(*args, **kwargs)
        self.update_fields()

    def update_fields(self):
        fields = []

        if hasattr(self.meta, 'fields'):
            for field_name in self.meta.fields:
                assert hasattr(self, field_name), (
                    'No such field %s for serializer %s' % (field_name, self.__class__.__name__)
                )
                field = getattr(self, field_name)
                field.field_name = field_name
                fields.append(field)

        if hasattr(self.meta, 'dynamic_fields'):
            for field_name, field in self.meta.dynamic_fields.items():
                field.field_name = field_name
                fields.append(field)

        for field_name in dir(self):
            field = getattr(self, field_name)

            if not isinstance(field, Field):
                continue

            field.field_name = field_name
            fields.append(field)

        self.fields = fields

    @property
    def readable_fields(self):
        if not self.fields or not isinstance(self.fields, Iterable):
            return []
        return list(filter(lambda x: not x.write_only, self.fields))

    @property
    def writable_fields(self):
        if not self.fields or not isinstance(self.fields, Iterable):
            return []
        return list(filter(lambda x: not x.read_only, self.fields))

    def run_validation(self, value):
        value = self.to_internal_value(value)
        return value

    def is_valid(self, raise_exception=False):
        try:
            self.validated_data = self.run_validation(self.data)
            self.errors = None
        except ValidationError as e:
            self.validated_data = None
            self.errors = e.detail

        if self.errors and raise_exception:
            raise ValidationError(self.errors)

        return not bool(self.errors)

    def to_internal_value_item(self, value):
        result = OrderedDict()
        errors = OrderedDict()

        for field in self.writable_fields:
            field_value = field.get_value(value)

            if field.field_name not in self.data.keys() and self.partial:
                continue

            validate_method = getattr(self, 'validate_' + field.field_name, None)

            try:
                validated_value = field.run_validation(field_value)
                if validate_method is not None:
                    validated_value = validate_method(validated_value)
                result[field.field_name] = validated_value
            except ValidationError as e:
                errors[field.field_name] = e.detail

        if errors:
            raise ValidationError(errors)

        return result

    def to_representation_item(self, value):
        result = OrderedDict()

        for field in self.readable_fields:
            if isinstance(value, Mapping):
                field_value = value[field.field_name]
            else:
                field_value = getattr(value, field.field_name)

            result[field.field_name] = field.to_representation(field_value)

        return result

    @property
    def representation_data(self):
        if self.instance is not None:
            return self.to_representation(self.instance)
        elif self.validated_data is not None:
            return self.to_representation(self.validated_data)

    def update(self, instance, validated_data):
        raise NotImplementedError('`update()` must be implemented.')

    def create(self, validated_data):
        raise NotImplementedError('`create()` must be implemented.')

    def save(self, **kwargs):
        assert not self.errors, (
            'You cannot call `.save()` on a serializer with invalid data.'
        )

        validated_data = dict(
            list(self.validated_data.items()) +
            list(kwargs.items())
        )

        if self.instance is not None:
            self.instance = self.update(self.instance, validated_data)
            assert self.instance is not None, (
                '`update()` did not return an object instance.'
            )
        else:
            self.instance = self.create(validated_data)
            assert self.instance is not None, (
                '`create()` did not return an object instance.'
            )

        return self.instance



from datetime import datetime

import logging
import requests

from jet_bridge import settings
from jet_bridge.db import Session
from jet_bridge.models.token import Token


def api_method_url(method):
    return '{}/{}'.format(settings.API_BASE_URL, method)


def register_token():
    try:
        session = Session()
        token = session.query(Token).first()

        if token:
            return token, False

        url = api_method_url('project_tokens/')
        headers = {
            'User-Agent': 'Jet Django'
        }

        r = requests.request('POST', url, headers=headers)
        success = 200 <= r.status_code < 300

        if not success:
            logging.error('Register Token request error', r.status_code, r.reason)
            return None, False

        result = r.json()

        # TODO: add serializer
        token = result['token'].replace('-', '')
        date_add = datetime.strptime(result['date_add'][:-6], '%Y-%m-%dT%H:%M:%S.%f')

        token = Token(token=token, date_add=date_add)
        session.add(token)
        session.commit()

        return token, True
    finally:
        session.close()


def is_token_activated():
    try:
        session = Session()
        token = session.query(Token).first()

        if not token:
            return False

        url = api_method_url('project_tokens/{}/'.format(token.token))
        headers = {
            'User-Agent': 'Jet Django'
        }

        r = requests.request('GET', url, headers=headers)
        success = 200 <= r.status_code < 300

        if not success:
            return False

        result = r.json()

        return bool(result.get('activated'))
    finally:
        session.close()


def reset_token():
    try:
        session = Session()
        session.query(Token).delete()
        session.commit()

        return register_token()
    finally:
        session.close()


def project_auth(token, permission=None):
    try:
        session = Session()
        project_token = session.query(Token).first()

        if not project_token:
            return {
                'result': False
            }

        url = api_method_url('project_auth/')
        data = {
            'project_token': project_token.token,
            'token': token
        }
        headers = {
            'User-Agent': 'Jet Django'
        }

        if permission:
            data.update(permission)

        r = requests.request('POST', url, data=data, headers=headers)
        success = 200 <= r.status_code < 300

        if not success:
            logging.error('Project Auth request error', r.status_code, r.reason)
            return {
                'result': False
            }

        result = r.json()

        if result.get('access_disabled'):
            return {
                'result': False,
                'warning': result.get('warning')
            }

        return {
            'result': True,
            'warning': result.get('warning')
        }
    finally:
        session.close()



from sqlalchemy import inspect

from jet_bridge import fields
from jet_bridge.serializers.serializer import Serializer


def get_reorder_serializer(Model, queryset, session):
    class ReorderSerializer(Serializer):
        ordering_field = fields.CharField()
        forward = fields.BooleanField()
        segment_from = fields.IntegerField()
        segment_to = fields.IntegerField()
        item = fields.IntegerField()
        segment_by_ordering_field = fields.BooleanField(default=False)

        def save(self):
            mapper = inspect(Model)

            primary_key_field = mapper.primary_key[0].name
            ordering_field = self.validated_data['ordering_field']
            primary_key = getattr(Model, primary_key_field)
            ordering = getattr(Model, ordering_field)

            if self.validated_data.get('segment_by_ordering_field'):
                segment_from = self.validated_data['segment_from']
                segment_to = self.validated_data['segment_to']
            else:
                segment_from_instance = queryset.filter(primary_key == self.validated_data['segment_from']).first()
                segment_to_instance = queryset.filter(primary_key == self.validated_data['segment_to']).first()

                segment_from = getattr(segment_from_instance, ordering_field)
                segment_to = getattr(segment_to_instance, ordering_field)

            if self.validated_data['forward']:
                queryset.filter(
                    ordering >= segment_from,
                    ordering <= segment_to
                ).update(
                    {ordering_field: ordering - 1}
                )
                queryset.filter(
                    primary_key == self.validated_data['item']
                ).update(
                    {ordering_field: segment_to}
                )
            else:
                queryset.filter(
                    ordering >= segment_from,
                    ordering <= segment_to
                ).update(
                    {ordering_field: ordering + 1}
                )
                queryset.filter(
                    primary_key == self.validated_data['item']
                ).update(
                    {ordering_field: segment_to}
                )

            session.commit()

    return ReorderSerializer



from sqlalchemy import inspect, desc
from sqlalchemy.sql import operators
from sqlalchemy.sql.elements import UnaryExpression, AnnotatedColumnElement

from jet_bridge.filters.model import get_model_filter_class
from jet_bridge.filters.model_aggregate import ModelAggregateFilter
from jet_bridge.filters.model_group import ModelGroupFilter
from jet_bridge.permissions import HasProjectPermissions, ModifyNotInDemo
from jet_bridge.responses.base import Response
from jet_bridge.router import action
from jet_bridge.serializers.model import get_model_serializer
from jet_bridge.serializers.model_group import ModelGroupSerializer
from jet_bridge.serializers.reorder import get_reorder_serializer
from jet_bridge.serializers.reset_order import get_reset_order_serializer
from jet_bridge.utils.siblings import get_model_siblings
from jet_bridge.views.mixins.model import ModelAPIViewMixin
from jet_bridge.db import MappedBase


class ModelHandler(ModelAPIViewMixin):
    model = None
    permission_classes = (HasProjectPermissions, ModifyNotInDemo)

    def prepare(self):
        super(ModelHandler, self).prepare()
        mapper = inspect(self.get_model())
        self.lookup_field = mapper.primary_key[0].name

    @property
    def required_project_permission(self):
        return {
            'permission_type': 'model',
            'permission_object': self.path_kwargs['model'],
            'permission_actions': {
                'create': 'w',
                'update': 'w',
                'partial_update': 'w',
                'destroy': 'd',
                'retrieve': 'r',
                'list': 'r',
                'aggregate': 'r',
                'group': 'r',
                'reorder': 'w',
                'reset_order': 'w'
            }.get(self.action, 'w')
        }

    def get_model(self):
        if self.model:
            return self.model

        self.model = MappedBase.classes[self.path_kwargs['model']]

        return self.model

    def get_serializer_class(self):
        Model = self.get_model()
        return get_model_serializer(Model)

    def get_filter_class(self):
        return get_model_filter_class(self.get_model())

    def get_queryset(self):
        Model = self.get_model()

        return self.session.query(Model)

    def filter_queryset(self, queryset):
        queryset = super().filter_queryset(queryset)
        if self.action == 'list':
            mapper = inspect(self.model)
            pk = mapper.primary_key[0].name
            context = queryset._compile_context()
            ordering = context.order_by

            def is_pk(x):
                if isinstance(x, AnnotatedColumnElement):
                    return x.name == pk
                elif isinstance(x, UnaryExpression):
                    return x.element.name == pk and x.modifier == operators.desc_op
                return False

            if ordering is None or not any(map(is_pk, ordering)):
                order_by = list(ordering or []) + [desc(pk)]
                queryset = queryset.order_by(*order_by)
        return queryset

    @action(methods=['get'], detail=False)
    def aggregate(self, *args, **kwargs):
        queryset = self.filter_queryset(self.get_queryset())

        y_func = self.get_argument('_y_func').lower()
        y_column = self.get_argument('_y_column', self.lookup_field)

        model_serializer = self.get_serializer()

        y_serializers = list(filter(lambda x: x.field_name == y_column, model_serializer.fields))
        y_serializer = y_serializers[0]

        filter_instance = ModelAggregateFilter()
        filter_instance.model = self.model
        queryset = filter_instance.filter(queryset, {
            'y_func': y_func,
            'y_column': y_column
        })

        result = y_serializer.to_representation(queryset[0])  # TODO: Refactor serializer

        self.write_response(Response({
            'y_func': result
        }))

    @action(methods=['get'], detail=False)
    def group(self, *args, **kwargs):
        queryset = self.filter_queryset(self.get_queryset())

        x_column = self.get_argument('_x_column')
        x_lookup_name = self.get_argument('_x_lookup', None)
        y_func = self.get_argument('_y_func').lower()
        y_column = self.get_argument('_y_column', self.lookup_field)

        model_serializer = self.get_serializer()

        # x_serializers = list(filter(lambda x: x.field_name == x_column, model_serializer.fields))
        # x_serializer = x_serializers[0]

        # y_serializers = list(filter(lambda x: x.field_name == y_column, model_serializer.fields))
        # y_serializer = y_serializers[0]

        filter_instance = ModelGroupFilter()
        filter_instance.model = self.model
        queryset = filter_instance.filter(queryset, {
            'x_column': x_column,
            'x_lookup': x_lookup_name,
            'y_func': y_func,
            'y_column': y_column
        })
        serializer = ModelGroupSerializer(
            instance=queryset,
            many=True,
            # TODO: Refactor serializer
            # group_serializer=x_serializer,
            # y_func_serializer=y_serializer
        )

        self.write_response(Response(serializer.representation_data))

    @action(methods=['post'], detail=False)
    def reorder(self, *args, **kwargs):
        queryset = self.filter_queryset(self.get_queryset())
        ReorderSerializer = get_reorder_serializer(self.get_model(), queryset, self.session)

        serializer = ReorderSerializer(data=self.data)
        serializer.is_valid(raise_exception=True)
        serializer.save()

        self.write_response(Response(serializer.representation_data))

    @action(methods=['post'], detail=False)
    def reset_order(self, *args, **kwargs):
        queryset = self.filter_queryset(self.get_queryset())
        ResetOrderSerializer = get_reset_order_serializer(self.get_model(), queryset, self.session)

        serializer = ResetOrderSerializer(data=self.data)
        serializer.is_valid(raise_exception=True)
        serializer.save()

        self.write_response(Response(serializer.representation_data))

    @action(methods=['get'], detail=True)
    def get_siblings(self, *args, **kwargs):
        lookup_url_kwarg = self.lookup_url_kwarg or 'pk'

        assert lookup_url_kwarg in self.path_kwargs

        model_field = getattr(self.get_model(), self.lookup_field)
        obj = self.get_queryset().filter(getattr(model_field, '__eq__')(self.path_kwargs[lookup_url_kwarg])).first()

        self.check_object_permissions(obj)

        queryset = self.filter_queryset(self.get_queryset())

        self.write_response(Response(get_model_siblings(self.model, obj, queryset)))



import logging

from sqlalchemy import create_engine, MetaData
from sqlalchemy.ext.automap import automap_base
from sqlalchemy.orm import sessionmaker

try:
    from geoalchemy2 import types
except ImportError:
    pass

from jet_bridge import settings
from jet_bridge.models import Base


def build_engine_url():
    if not settings.DATABASE_ENGINE or not settings.DATABASE_NAME:
        return

    url = [
        settings.DATABASE_ENGINE,
        '://'
    ]

    if settings.DATABASE_USER:
        url.append(settings.DATABASE_USER)

        if settings.DATABASE_PASSWORD:
            url.append(':')
            url.append(settings.DATABASE_PASSWORD)

        if settings.DATABASE_HOST:
            url.append('@')

    if settings.DATABASE_HOST:
        url.append(settings.DATABASE_HOST)

        if settings.DATABASE_PORT:
            url.append(':')
            url.append(str(settings.DATABASE_PORT))

        url.append('/')

    url.append(settings.DATABASE_NAME)

    return ''.join(url)

engine_url = build_engine_url()
Session = None

if engine_url:
    engine = create_engine(engine_url, pool_size=settings.CONNECTIONS, max_overflow=10)
    Session = sessionmaker(bind=engine)

    logging.info('Connected to database engine "{}" with name "{}"'.format(settings.DATABASE_ENGINE, settings.DATABASE_NAME))

    Base.metadata.create_all(engine)

    metadata = MetaData()
    metadata.reflect(engine)
    MappedBase = automap_base(metadata=metadata)
    MappedBase.prepare()



from jet_bridge.fields import CharField
from jet_bridge.serializers.serializer import Serializer


class ModelGroupSerializer(Serializer):
    group = CharField()
    y_func = CharField()  # TODO: change to integer default

    def __init__(self, *args, **kwargs):
        if 'group_serializer' in kwargs:
            self.fields['group'] = kwargs.pop('group_serializer')

        if 'y_func_serializer' in kwargs:
            self.fields['y_func'] = kwargs.pop('y_func_serializer')

        super(ModelGroupSerializer, self).__init__(*args, **kwargs)

    class Meta:
        fields = (
            'group',
            'y_func',
        )



from sqlalchemy.sql import sqltypes

from jet_bridge.filters import lookups
from jet_bridge.filters.boolean_filter import BooleanFilter
from jet_bridge.filters.char_filter import CharFilter
from jet_bridge.filters.datetime_filter import DateTimeFilter
from jet_bridge.filters.wkt_filter import WKTFilter
from jet_bridge.filters.integer_filter import IntegerFilter

number_lookups = [
    lookups.EXACT,
    lookups.GT,
    lookups.GTE,
    lookups.LT,
    lookups.LTE,
    lookups.ICONTAINS,
    lookups.IN,
    lookups.IS_NULL,
]

datetime_lookups = [
    lookups.EXACT,
    lookups.GT,
    lookups.GTE,
    lookups.LT,
    lookups.LTE,
    lookups.ICONTAINS,
    lookups.IN,
    lookups.IS_NULL,
]

text_lookups = [
    lookups.EXACT,
    lookups.ICONTAINS,
    lookups.IN,
    lookups.STARTS_WITH,
    lookups.ENDS_WITH,
    lookups.IS_NULL,
]

boolean_lookups = [
    lookups.EXACT,
    lookups.IN,
    lookups.IS_NULL,
]

json_lookups = [
    lookups.JSON_ICONTAINS,
    lookups.IS_NULL,
]

geography_lookups = [
    lookups.COVEREDBY
]

FILTER_FOR_DBFIELD = {
    sqltypes.VARCHAR: {'filter_class': CharFilter, 'lookups': text_lookups},
    sqltypes.TEXT: {'filter_class': CharFilter, 'lookups': text_lookups},
    sqltypes.BOOLEAN: {'filter_class': BooleanFilter, 'lookups': boolean_lookups},
    sqltypes.INTEGER: {'filter_class': IntegerFilter, 'lookups': number_lookups},
    sqltypes.SMALLINT: {'filter_class': IntegerFilter, 'lookups': number_lookups},
    sqltypes.NUMERIC: {'filter_class': IntegerFilter, 'lookups': number_lookups},
    sqltypes.DATETIME: {'filter_class': DateTimeFilter, 'lookups': datetime_lookups},
    sqltypes.TIMESTAMP: {'filter_class': DateTimeFilter, 'lookups': datetime_lookups},
    sqltypes.JSON: {'filter_class': CharFilter, 'lookups': json_lookups},
    # sqlalchemy.TextField:                   {'filter_class': CharFilter},
    # sqlalchemy.BOOLEAN:                {'filter_class': BooleanFilter},
    # sqlalchemy.DateField:                   {'filter_class': DateFilter},
    # sqlalchemy.DateTimeField:               {'filter_class': DateTimeFilter},
    # sqlalchemy.TimeField:                   {'filter_class': TimeFilter},
    # sqlalchemy.DurationField:               {'filter_class': DurationFilter},
    # sqlalchemy.DecimalField:                {'filter_class': NumberFilter},
    # sqlalchemy.SmallIntegerField:           {'filter_class': NumberFilter},
    # sqlalchemy.IntegerField:                {'filter_class': NumberFilter},
    # sqlalchemy.PositiveIntegerField:        {'filter_class': NumberFilter},
    # sqlalchemy.PositiveSmallIntegerField:   {'filter_class': NumberFilter},
    # sqlalchemy.FloatField:                  {'filter_class': NumberFilter},
    # sqlalchemy.NullBooleanField:            {'filter_class': BooleanFilter},
    # sqlalchemy.SlugField:                   {'filter_class': CharFilter},
    # sqlalchemy.EmailField:                  {'filter_class': CharFilter},
    # sqlalchemy.FilePathField:               {'filter_class': CharFilter},
    # sqlalchemy.URLField:                    {'filter_class': CharFilter},
    # sqlalchemy.GenericIPAddressField:       {'filter_class': CharFilter},
    # sqlalchemy.CommaSeparatedIntegerField:  {'filter_class': CharFilter},
    # sqlalchemy.UUIDField:                   {'filter_class': UUIDFilter}
}
FILTER_FOR_DBFIELD_DEFAULT = FILTER_FOR_DBFIELD[sqltypes.VARCHAR]

try:
    from geoalchemy2 import types
    FILTER_FOR_DBFIELD[types.Geometry] = {'filter_class': WKTFilter, 'lookups': geography_lookups}
    FILTER_FOR_DBFIELD[types.Geography] = {'filter_class': WKTFilter, 'lookups': geography_lookups}
except ImportError:
    pass


def filter_for_data_type(value):
    for date_type, filter_data in FILTER_FOR_DBFIELD.items():
        if isinstance(value, date_type):
            return filter_data
    return FILTER_FOR_DBFIELD_DEFAULT



from jet_bridge import status
from jet_bridge.responses.base import Response


class CreateAPIViewMixin(object):

    def create(self, *args, **kwargs):
        serializer = self.get_serializer(data=self.data)
        serializer.is_valid(raise_exception=True)
        self.perform_create(serializer)
        self.write_response(Response(serializer.representation_data, status=status.HTTP_201_CREATED))

    def perform_create(self, serializer):
        serializer.save()



from datetime import datetime

import sys
import platform
import tornado.web
from tornado.escape import json_decode

from jet_bridge import settings, VERSION, status
from jet_bridge.db import Session
from jet_bridge.exceptions.api import APIException
from jet_bridge.exceptions.permission_denied import PermissionDenied


class APIView(tornado.web.RequestHandler):
    permission_classes = []
    session = None

    @property
    def data(self):
        content_type = self.request.headers.get('Content-Type', '')
        if content_type.startswith('application/json'):
            return json_decode(self.request.body)
        else:
            return self.request.body_arguments

    def prepare(self):
        method_override = self.request.headers.get('X-Http-Method-Override')
        if method_override is not None:
            self.request.method = method_override

        if self.request.method != 'OPTIONS':
            self.check_permissions()

        self.session = Session()

    def on_finish(self):
        if self.session:
            self.session.close()
            self.session = None

    def set_default_headers(self):
        ACCESS_CONTROL_ALLOW_ORIGIN = 'Access-Control-Allow-Origin'
        ACCESS_CONTROL_EXPOSE_HEADERS = 'Access-Control-Expose-Headers'
        ACCESS_CONTROL_ALLOW_CREDENTIALS = 'Access-Control-Allow-Credentials'
        ACCESS_CONTROL_ALLOW_HEADERS = 'Access-Control-Allow-Headers'
        ACCESS_CONTROL_ALLOW_METHODS = 'Access-Control-Allow-Methods'

        self.set_header(ACCESS_CONTROL_ALLOW_ORIGIN, '*')
        self.set_header(ACCESS_CONTROL_ALLOW_METHODS, 'GET, POST, PUT, PATCH, DELETE, OPTIONS')
        self.set_header(ACCESS_CONTROL_ALLOW_HEADERS, 'Authorization,DNT,User-Agent,X-Requested-With,If-Modified-Since,Cache-Control,Content-Type,Range,X-Application-Warning,X-HTTP-Method-Override')
        self.set_header(ACCESS_CONTROL_EXPOSE_HEADERS, 'Content-Length,Content-Range,X-Application-Warning')
        self.set_header(ACCESS_CONTROL_ALLOW_CREDENTIALS, 'true')

    def get_permissions(self):
        return [permission() for permission in self.permission_classes]

    def check_permissions(self):
        for permission in self.get_permissions():
            if not permission.has_permission(self):
                raise PermissionDenied(getattr(permission, 'message', 'forbidden'))

    def check_object_permissions(self, obj):
        for permission in self.get_permissions():
            if not permission.has_object_permission(self, obj):
                raise PermissionDenied(getattr(permission, 'message', 'forbidden'))

    def options(self, *args, **kwargs):
        self.set_status(204)
        self.finish()

    def build_absolute_uri(self, url):
        return self.request.protocol + "://" + self.request.host + url

    def write_response(self, response):
        for name, value in response.header_items():
            self.set_header(name, value)
        self.write(response.render())

    def write_error(self, status_code, **kwargs):
        exc_type = exc = traceback = None

        if kwargs.get('exc_info'):
            exc_type, exc, traceback = kwargs['exc_info']

            if isinstance(exc, APIException):
                status_code = exc.status_code

        if status_code == status.HTTP_403_FORBIDDEN:
            self.render('403.html', **{
                'path': self.request.path,
            })
        else:
            if settings.DEBUG:
                ctx = {
                    'path': self.request.path,
                    'full_path':  self.request.protocol + "://" + self.request.host + self.request.path,
                    'method': self.request.method,
                    'version': VERSION,
                    'current_datetime': datetime.now().strftime('%c'),
                    'python_version': platform.python_version(),
                    'python_executable': sys.executable,
                    'python_path': sys.path
                }

                if exc:
                    last_traceback = traceback

                    while last_traceback.tb_next:
                        last_traceback = last_traceback.tb_next

                    ctx.update({
                        'exception_type': exc_type.__name__,
                        'exception_value': str(exc),
                        'exception_last_traceback_line': last_traceback.tb_lineno,
                        'exception_last_traceback_name': last_traceback.tb_frame
                    })

                self.render('500.debug.html', **ctx)
            else:
                self.render('500.html')



from jet_bridge.permissions import HasProjectPermissions
from jet_bridge.views.base.api import APIView


class MessageHandler(APIView):
    permission_classes = (HasProjectPermissions,)

    def post(self):
        pass



from jet_bridge.models import data_types

from sqlalchemy.sql import sqltypes

map_data_types = [
    {'query': sqltypes.VARCHAR, 'date_type': data_types.TEXT},
    {'query': sqltypes.TEXT, 'date_type': data_types.TEXT},
    {'query': sqltypes.BOOLEAN, 'date_type': data_types.BOOLEAN},
    {'query': sqltypes.INTEGER, 'date_type': data_types.INTEGER},
    {'query': sqltypes.SMALLINT, 'date_type': data_types.INTEGER},
    {'query': sqltypes.NUMERIC, 'date_type': data_types.FLOAT},
    {'query': sqltypes.DATETIME, 'date_type': data_types.DATE_TIME},
    {'query': sqltypes.TIMESTAMP, 'date_type': data_types.DATE_TIME},
    {'query': sqltypes.JSON, 'date_type': data_types.JSON},
]
default_data_type = data_types.TEXT

try:
    from geoalchemy2 import types
    map_data_types.append({'query': types.Geometry, 'date_type': data_types.GEOMETRY})
    map_data_types.append({'query': types.Geography, 'date_type': data_types.GEOGRAPHY})
except ImportError:
    pass


def map_data_type(value):
    for rule in reversed(map_data_types):
        if isinstance(value, rule['query']):
            return rule['date_type']
    return default_data_type




EXACT = 'exact'
GT = 'gt'
GTE = 'gte'
LT = 'lt'
LTE = 'lte'
ICONTAINS = 'icontains'
IN = 'in'
STARTS_WITH = 'istartswith'
ENDS_WITH = 'iendswith'
IS_NULL = 'isnull'
JSON_ICONTAINS = 'json_icontains'
COVEREDBY = 'coveredby'
DEFAULT_LOOKUP = EXACT



from jet_bridge.responses.base import Response


class UpdateAPIViewMixin(object):

    def update(self, *args, **kwargs):
        partial = kwargs.pop('partial', False)
        instance = self.get_object()
        serializer = self.get_serializer(instance=instance, data=self.data, partial=partial)
        serializer.is_valid(raise_exception=True)
        self.perform_update(serializer)
        self.write_response(Response(serializer.representation_data))

    def put(self, *args, **kwargs):
        self.update(*args, **kwargs)

    def patch(self, *args, **kwargs):
        self.update(partial=True, *args, **kwargs)

    def perform_update(self, serializer):
        serializer.save()



from sqlalchemy import Unicode
from sqlalchemy.dialects.postgresql import JSONB

from jet_bridge.fields import field, CharField, BooleanField
from jet_bridge.filters import lookups

EMPTY_VALUES = ([], (), {}, '', None)


def json_icontains(qs, model, field_name, value):
    model_field = getattr(model, field_name)
    field_type = model_field.property.columns[0].type

    if isinstance(field_type, JSONB):
        return qs.filter(model_field.cast(Unicode).ilike('%{}%'.format(value)))
    else:
        return qs.filter(model_field.astext.ilike('%{}%'.format(value)))


def coveredby(qs, model, field_name, value):
    model_field = getattr(model, field_name)
    return qs.filter(model_field.ST_CoveredBy(value))


class Filter(object):
    field_class = field
    lookup_operators = {
        lookups.EXACT: {'operator': '__eq__'},
        lookups.GT: {'operator': '__gt__'},
        lookups.GTE: {'operator': '__ge__'},
        lookups.LT: {'operator': '__lt__'},
        lookups.LTE: {'operator': '__le__'},
        lookups.ICONTAINS: {'operator': 'ilike', 'post_process': lambda x: '%{}%'.format(x)},
        lookups.IN: {'operator': 'in_', 'field_class': CharField, 'field_kwargs': {'many': True}, 'pre_process': lambda x: x.split(',')},
        lookups.STARTS_WITH: {'operator': 'ilike', 'post_process': lambda x: '{}%'.format(x)},
        lookups.ENDS_WITH: {'operator': 'ilike', 'post_process': lambda x: '%{}'.format(x)},
        lookups.IS_NULL: {'operator': lambda x: ('__eq__', None) if x else ('isnot', None), 'field_class': BooleanField},
        lookups.JSON_ICONTAINS: {'operator': False, 'func': json_icontains},
        lookups.COVEREDBY: {'operator': False, 'func': coveredby}
    }

    def __init__(self, field_name=None, model=None, lookup=lookups.DEFAULT_LOOKUP, request=None, handler=None):
        self.field_name = field_name
        self.name = field_name
        self.model = model
        self.lookup = lookup
        self.request = request
        self.handler = handler

    def clean_value(self, value):
        return value

    def apply_lookup(self, qs, lookup, value):
        model_field = getattr(self.model, self.field_name)
        lookup_operator = self.lookup_operators[lookup]
        operator = lookup_operator['operator']
        pre_process = lookup_operator.get('pre_process')
        post_process = lookup_operator.get('post_process')
        field_class = lookup_operator.get('field_class')
        field_kwargs = lookup_operator.get('field_kwargs', {})
        func = lookup_operator.get('func')

        if pre_process:
            value = pre_process(value)

        if field_class:
            value = field_class(**field_kwargs).to_internal_value(value)

        if post_process:
            value = post_process(value)

        if func:
            return func(qs, self.model, self.field_name, value)
        elif callable(operator):
            op = operator(value)
            return qs.filter(getattr(model_field, op[0])(op[1]))
        else:
            return qs.filter(getattr(model_field, operator)(value))

    def filter(self, qs, value):
        value = self.clean_value(value)
        if value in EMPTY_VALUES:
            return qs
        return self.apply_lookup(qs, self.lookup, value)



import os

from prompt_toolkit import prompt, print_formatted_text, HTML
from prompt_toolkit.completion import WordCompleter
from prompt_toolkit.validation import Validator

from jet_bridge import settings


engines = [
    {
        'name': 'postgresql',
        'default_port': 5432
    },
    {
        'name': 'mysql',
        'default_port': 3306
    },
    {
        'name': 'oracle',
        'default_port': 1521
    },
    {
        'name': 'mssql',
        'default_port': 1433
    },
    {
        'name': 'sqlite'
    }
]


def is_not_empty(text):
    return text is not None and text.strip() != ''


def database_engine_is_valid(text):
    return text in list(map(lambda x: x['name'], engines))


def port_is_valid(text):
    try:
        int(text)
        return True
    except ValueError as e:
        return False


def is_file_exists(text):
    return os.path.exists(text) and os.path.isfile(text)


promt_messages = 0


def create_config():
    print_formatted_text(HTML('<skyblue><b>===========================================</b></skyblue>'))
    print_formatted_text(HTML('<skyblue><b>Configuration file is not found</b></skyblue>'))
    print_formatted_text(HTML('<skyblue>You will be asked to enter settings</skyblue>'))
    print_formatted_text(HTML('<skyblue>and config file will be generated</skyblue>'))
    print_formatted_text(HTML('<skyblue><b>===========================================</b></skyblue>\n'))

    # session = PromptSession()
    html_completer = WordCompleter(list(map(lambda x: x['name'], engines)))

    def promt_message(message):
        global promt_messages
        promt_messages += 1
        return HTML('{}. {}\n> '.format(promt_messages, message))

    address = prompt(
        promt_message('<green><b>Which host to run Jet Bridge on?</b></green>\n<i>Default is {}</i>'.format('0.0.0.0 (any IP)')),
        default='0.0.0.0'
    )

    print_formatted_text('')

    port = prompt(
        promt_message('<green><b>Which port?</b></green>\n<i>Default is {}</i>'.format('8888')),
        validator=Validator.from_callable(
            port_is_valid,
            error_message='Incorrect port',
            move_cursor_to_end=True
        ),
        default='8888'
    )

    print_formatted_text('')

    database_engine = prompt(
        promt_message('<green><b>Which database do you use?</b></green>\n<i>Should be one of: {}</i>'.format(
            ', '.join(map(lambda x: x['name'], engines)))),
        validator=Validator.from_callable(
            database_engine_is_valid,
            error_message='Unknown database engine',
            move_cursor_to_end=True
        ),
        completer=html_completer,
        validate_while_typing=False
    )

    print_formatted_text('')

    if database_engine == 'sqlite':
        database_name = prompt(
            promt_message('<green><b>Enter your database file path</b></green>'),
            validator=Validator.from_callable(
                is_file_exists,
                error_message='File does not exist on given path',
                move_cursor_to_end=True
            )
        )

        database_host = ''
        database_port = ''
        database_user = ''
        database_password = ''
    else:
        database_host = prompt(
            promt_message('<green><b>Enter your database host</b></green>\n<i>Default is {}</i>'.format('localhost')),
            validator=Validator.from_callable(
                is_not_empty,
                error_message='Database host is required',
                move_cursor_to_end=True
            ),
            default='localhost'
        )

        print_formatted_text('')

        default_port = list(map(lambda x: str(x.get('default_port', '')), filter(
            lambda x: x['name'] == database_engine, engines
        )))[0]

        database_port = prompt(
            promt_message('<green><b>Enter your database port</b></green>'),
            validator=Validator.from_callable(
                is_not_empty,
                error_message='Database port is required',
                move_cursor_to_end=True
            ),
            default=default_port
        )

        print_formatted_text('')

        database_name = prompt(
            promt_message('<green><b>Enter your database name</b></green>'),
            validator=Validator.from_callable(
                is_not_empty,
                error_message='Database name is required',
                move_cursor_to_end=True
            )
        )

        print_formatted_text('')

        database_user = prompt(
            promt_message('<green><b>Enter your database user</b></green>'),
            validator=Validator.from_callable(
                is_not_empty,
                error_message='Database user is required',
                move_cursor_to_end=True
            )
        )

        print_formatted_text('')

        database_password = prompt(
            promt_message('<green><b>Enter your database password</b></green>')
        )

    print_formatted_text('')

    config = prompt(
        promt_message('<green><b>Where to store config file?</b></green>\nDefault is jet.conf inside current working directory\n[{}]'.format(os.path.abspath(settings.DEFAULT_CONFIG_PATH))),
        default=settings.DEFAULT_CONFIG_PATH
    )

    config_content = {
        'ADDRESS': address,
        'PORT': port,
        'CONFIG': config,
        'DATABASE_ENGINE': database_engine,
        'DATABASE_HOST': database_host,
        'DATABASE_PORT': database_port,
        'DATABASE_NAME': database_name,
        'DATABASE_USER': database_user,
        'DATABASE_PASSWORD': database_password
    }

    with open(config, 'w') as f:
        f.write('[JET]\n')
        for key, value in config_content.items():
            f.write('{}={}\n'.format(key, value))

    print_formatted_text('')
    print_formatted_text(HTML('<skyblue><b>===========================================</b></skyblue>'))
    print_formatted_text(HTML('<skyblue><b>Configuration file is created at:</b></skyblue>'))
    print_formatted_text(HTML('<skyblue>{}</skyblue>'.format(os.path.abspath(config))))
    print_formatted_text('')
    print_formatted_text(HTML('<skyblue>You can now start <b>Jet Bridge</b> by running</skyblue>'))
    print_formatted_text(HTML('<skyblue><b>jet_bridge</b> command once again</skyblue>'))
    print_formatted_text(HTML('<skyblue><b>===========================================</b></skyblue>\n'))



import os
from setuptools import setup, find_packages


def read(fname):
    path = os.path.join(os.path.dirname(__file__), fname)
    try:
        file = open(path, encoding='utf-8')
    except TypeError:
        file = open(path)
    return file.read()


def get_install_requires():
    install_requires = [
        'tornado',
        'sqlalchemy',
        'six',
        'requests',
        'Pillow',
        'dateparser',
        'prompt_toolkit>=2.0.9',
    ]

    try:
        from collections import OrderedDict
    except ImportError:
        install_requires.append('ordereddict')

    return install_requires

setup(
    name='jet_bridge',
    version=__import__('jet_bridge').VERSION,
    author='Denis Kildishev',
    author_email='support@jetadmin.io',
    url='https://github.com/jet-admin/jet-bridge',
    packages=find_packages(),
    license='MIT',
    classifiers=[

    ],
    zip_safe=False,
    include_package_data=True,
    install_requires=get_install_requires(),
    entry_points={
        'console_scripts': [
            'jet_bridge = jet_bridge.__main__:main',
        ],
    },
)



from jet_bridge.fields.char import CharField
from jet_bridge.fields.boolean import BooleanField
from jet_bridge.fields.integer import IntegerField
from jet_bridge.fields.datetime import DateTimeField
from jet_bridge.fields.json import JSONField
from jet_bridge.fields.wkt import WKTField
from jet_bridge.fields.array import ArrayField



from datetime import datetime
import webbrowser

import logging

import tornado.ioloop
import tornado.web
from requests import RequestException

import jet_bridge.adapters.postgres

from jet_bridge import settings, VERSION
from jet_bridge.settings import missing_options
from jet_bridge.utils.backend import is_token_activated
from jet_bridge.utils.create_config import create_config

logging.getLogger().setLevel(logging.INFO)


def main():
    logging.info(datetime.now().strftime('%B %d, %Y - %H:%M:%S %Z'))
    logging.info('Jet Bridge version {}'.format(VERSION))

    if missing_options == settings.required_options_without_default:
        create_config()
        return
    elif len(missing_options) and len(missing_options) < len(settings.required_options_without_default):
        logging.info('Required options are not specified: {}'.format(', '.join(missing_options)))
        return

    from jet_bridge.app import make_app

    app = make_app()
    app.listen(settings.PORT, settings.ADDRESS)
    address = 'localhost' if settings.ADDRESS == '0.0.0.0' else settings.ADDRESS
    url = 'http://{}:{}/'.format(address, settings.PORT)

    logging.info('Starting server at {}'.format(url))

    if settings.DEBUG:
        logging.warning('Server is running in DEBUG mode')

    logging.info('Quit the server with CONTROL-C')

    try:
        if not is_token_activated():
            register_url = '{}api/register/'.format(url)
            logging.warning('[!] Your server token is not activated')

            if webbrowser.open(register_url):
                logging.warning('[!] Activation page was opened in your browser - {}'.format(register_url))
            else:
                logging.warning('[!] Go to {} to activate'.format(register_url))
    except RequestException:
        logging.error('[!] Can\'t connect to Jet Admin API')
        logging.error('[!] Token verification failed')

    tornado.ioloop.IOLoop.current().start()

if __name__ == '__main__':
    main()



from jet_bridge.fields.field import Field


class WKTField(Field):

    def to_internal_value_item(self, value):
        if value is None:
            return
        from geoalchemy2 import WKTElement
        return WKTElement(value)

    def to_representation_item(self, value):
        if value is None:
            return
        from geoalchemy2.shape import to_shape
        return to_shape(value).to_wkt()



from jet_bridge import status


class ErrorDetail(object):
    code = None

    def __init__(self, string, code=None):
        self.string = string
        self.code = code

    def __eq__(self, other):
        try:
            return self.code == other.code
        except AttributeError:
            return False

    def __ne__(self, other):
        return not self.__eq__(other)

    def __repr__(self):
        return 'ErrorDetail(string=%r, code=%r)' % (
            self.string,
            self.code,
        )


def _get_error_details(data, code=None):
    if isinstance(data, list):
        ret = [
            _get_error_details(item, code) for item in data
        ]
        return ret
    elif isinstance(data, dict):
        ret = {
            key: _get_error_details(value, code) for key, value in data.items()
        }
        return ret
    else:
        text = str(data)
        code = getattr(data, 'code', code)
        return ErrorDetail(text, code)


def _get_codes(detail):
    if isinstance(detail, list):
        return [_get_codes(item) for item in detail]
    elif isinstance(detail, dict):
        return {key: _get_codes(value) for key, value in detail.items()}
    return detail.code


def _get_full_details(detail):
    if isinstance(detail, list):
        return [_get_full_details(item) for item in detail]
    elif isinstance(detail, dict):
        return {key: _get_full_details(value) for key, value in detail.items()}
    return {
        'message': detail,
        'code': detail.code
    }


class APIException(Exception):
    default_detail = 'A server error occurred.'
    default_code = 'error'
    default_status_code = status.HTTP_500_INTERNAL_SERVER_ERROR

    def __init__(self, detail=None, code=None, status_code=None):
        if detail is None:
            detail = self.default_detail
        if code is None:
            code = self.default_code
        if status_code is None:
            status_code = self.default_status_code

        self.detail = _get_error_details(detail, code)
        self.code = code
        self.status_code = status_code

    def __str__(self):
        return str(self.detail)

    def get_codes(self):
        """
        Return only the code part of the error details.

        Eg. {"name": ["required"]}
        """
        return _get_codes(self.detail)

    def get_full_details(self):
        """
        Return both the message & code parts of the error details.

        Eg. {"name": [{"message": "This field is required.", "code": "required"}]}
        """
        return _get_full_details(self.detail)
